//
// This file is part of the aMule Project.
//
// Copyright (c) 2003-2006 aMule Team ( admin@amule.org / http://www.amule.org )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include <algorithm>
#include <cstring>
#include <cmath>

#include "Types.h"

#ifdef __WXMSW__
	#include <winsock.h>
	#include <wx/defs.h>
	#include <wx/msw/winundef.h>
#else
	#include <sys/types.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
#endif

#include "UploadQueue.h"	// Interface declarations
#include "ServerList.h"		// Needed for CServerList
#include "DownloadQueue.h"	// Needed for CDownloadQueue
#include "Server.h"		// Needed for CServer
#include "ServerConnect.h"	// Needed for CServerConnect
#include "KnownFile.h"		// Needed for CKnownFile
#include "Packet.h"		// Needed for CPacket
#include "ClientTCPSocket.h"	// Needed for CClientTCPSocket
#include "SharedFileList.h"	// Needed for CSharedFileList
#include "OPCodes.h"		// Needed for MAX_PURGEQUEUETIME
#include "updownclient.h"	// Needed for CUpDownClient
#include "GetTickCount.h"	// Needed for GetTickCount
#include "amule.h"		// Needed for theApp
#include "Preferences.h"
#include "ClientList.h"
#include "Statistics.h"		// Needed for theStats
#include "Logger.h"
#include <common/Format.h>
#include "UploadBandwidthThrottler.h"

// Mod adu


#include <numeric>

//TODO rewrite the whole networkcode, use overlapped sockets

CUploadQueue::CUploadQueue()
{
	m_nLastStartUpload = 0;

	lastupslotHighID = true;
	// mod Adu
	// Emanem
	// Inizializzo il numero di clients Adu
	m_AduClientsNum = 0;
	// fine mod Adu
}


void CUploadQueue::AddUpNextClient(CUpDownClient* directadd, uint32 tC)
{
	POSITION toadd = 0;
	POSITION toaddlow = 0;
	sint64	bestscore = -1;
	sint64	bestlowscore = -1;

	// Mod Adu
	// lupz
	// capisco cosa voglio
	uint32 typeClient = AduNextClient();
	//AddDebugLogLineM(false, logClient, CFormat(wxT("voglio: %u")) % typeClient);
	if (typeClient == ADUNANZA_NONE)
		return;
	// Fine mod Adu

	// Mod Adu
	// Emanem
	// Forziamo che il rapporto 1:1 sia rispettato
	//if (thePrefs.m_AduForceExt11 &&
	//	ADUNANZA_FASTWEB == AduGetTypeBand(0) &&
	//	0 == m_AduClientsNum)

	// lupz: aggiorno alla nuova logica

	// lupz: sistemo anche il criterio
	// il vecchio criterio causava frequenti coda piena

	if (directadd) {
		switch(typeClient) {
			case ADUNANZA_FASTWEB:
				if (!directadd->IsAduClient())
					return;
				break;
			case ADUNANZA_EXTERN:
				if (directadd->IsAduClient())
					return;
				break;
		}
	} else if (!directadd) {
		switch(typeClient) {
			case ADUNANZA_FASTWEB:
				if (!m_AduClientsNum)
					return;
				break;
			case ADUNANZA_EXTERN:
				if (m_AduClientsNum >= waitinglist.GetCount())
					return;
				break;
		}		
	}

	// fine mod Adu

	CUpDownClient* newclient;
	// select next client or use given client
	if (!directadd) {
		// mod Adu
		// Emanem
		// Risistemo il valore temporaneo del numero di clients
		// AdunanzA
		uint32 _tmp_m_AduClientsNum = 0;
		// fine mod Adu

		// Track if we purged any clients from the queue, as to only send one notify in total
		bool purged = false;
		
		POSITION pos1, pos2;
		for (pos1 = waitinglist.GetHeadPosition();( pos2 = pos1 ) != NULL;) {
			waitinglist.GetNext(pos1);
			CUpDownClient* cur_client = waitinglist.GetAt(pos2);
			// clear dead clients
			if ( (::GetTickCount() - cur_client->GetLastUpRequest() > MAX_PURGEQUEUETIME) || !theApp.sharedfiles->GetFileByID(cur_client->GetUploadFileID()) ) {
				purged = true;
				cur_client->ClearWaitStartTime();
				RemoveFromWaitingQueue(pos2);
				if (!cur_client->GetSocket()) {
					if(cur_client->Disconnected(wxT("AddUpNextClient - purged"))) {
						cur_client->Safe_Delete();
						cur_client = NULL;
					}
				}
				continue;
			} 

			suspendlist::iterator it = std::find( suspended_uploads_list.begin(),
			                                      suspended_uploads_list.end(),
			                                      cur_client->GetUploadFileID() );
			if (cur_client->IsBanned() || it != suspended_uploads_list.end() ) { // Banned client or suspended upload ?
			        continue;
			}
			// finished clearing

			// Codice AdunanzA
			// Emanem	18:15	4/4/2004
			//
			// In questo caso, in base al tipo di typeClient
			// in input si sceglie su che clients fare la "classifica"
			/// di priorita'
			uint32 cur_score = cur_client->GetScore(true);
		
			// valore temporaneo per sapere se un client e' AdunanzA
			// o no
			uint32 aduResult = cur_client->GetClientAduType();
			// nel caso incremento il contatore temporaneo
			if (cur_client->IsAduClient()) 
				_tmp_m_AduClientsNum++;
			// Se il tipo di client non e' disponibile
			// allora prendo il primo che mi capita
			if (typeClient == ADUNANZA_FASTWEB)
			{
				if (!m_AduClientsNum) typeClient = ADUNANZA_ANY;
			}
			if (typeClient == ADUNANZA_EXTERN)
			{
				if (m_AduClientsNum >= waitinglist.GetCount()) typeClient = ADUNANZA_ANY;
			}

			bool doit = false;
			switch(typeClient)
			{
				case ADUNANZA_EXTERN:
					doit = cur_client->IsAduClient() == false;
					break;
				case ADUNANZA_FASTWEB:
					doit = cur_client->IsAduClient() == true;
					break;
				default:
				case ADUNANZA_ANY:
					doit = true;
					break;
			};

			if (doit) {
				if ( cur_score > bestscore){
					bestscore = cur_score;
					toadd = pos2;
				} else {
					cur_score = cur_client->GetScore(false);
					if ((cur_score > bestlowscore) && !cur_client->m_bAddNextConnect){
						bestlowscore = cur_score;
						toaddlow = pos2;
					}
				}
			}
		}
		// mod Adu
		// Emanem
		// risistemo il numero di clients AdunanzA
		m_AduClientsNum = _tmp_m_AduClientsNum;

		// Update the count on GUI if any clients were purged
		if (purged) {
			Notify_ShowQueueCount(waitinglist.GetCount());
		}

		if (bestlowscore > bestscore){
			newclient = waitinglist.GetAt(toaddlow);
			newclient->m_bAddNextConnect = true;
		}

		if (!toadd) {
			return;
		}
		newclient = waitinglist.GetAt(toadd);
		lastupslotHighID = true; // VQB LowID alternate
		RemoveFromWaitingQueue(toadd);
		Notify_ShowQueueCount(waitinglist.GetCount());
	} else {
		//prevent another potential access of a suspended upload

		suspendlist::iterator it = std::find( suspended_uploads_list.begin(),
		                                      suspended_uploads_list.end(),
		                                      directadd->GetUploadFileID() );
		if ( it != suspended_uploads_list.end() ) {
			return;
		} else {
			newclient = directadd;
		}
	}

	if (IsDownloading(newclient)) {
		return;
	}
	// tell the client that we are now ready to upload
	if (!newclient->IsConnected()) {
		newclient->SetUploadState(US_CONNECTING);
		if (!newclient->TryToConnect(true)) {
			return;
		}
	} else {
		CPacket* packet = new CPacket(OP_ACCEPTUPLOADREQ,0);
		theStats::AddUpOverheadFileRequest(packet->GetPacketSize());
		newclient->SendPacket(packet,true);
		newclient->SetUploadState(US_UPLOADING);
	}
	newclient->SetUpStartTime();
	newclient->ResetSessionUp();

	theApp.uploadBandwidthThrottler->AddToStandardList(uploadinglist.GetCount(), newclient->GetSocket());
	uploadinglist.AddTail(newclient);
	// Mod Adu
	// lupz
		if (newclient->IsAduClient())
			aduuploadinglist.AddTail(newclient);
	// Fine mod Adu
	theStats::AddUploadingClient();

	// Statistic
	CKnownFile* reqfile = (CKnownFile*) newclient->GetUploadFile();
	if (reqfile) {
		reqfile->statistic.AddAccepted();
	}
	Notify_UploadCtrlAddClient(newclient);

}

void CUploadQueue::Process()
{
	if (AcceptNewClient(AduNextClient() == ADUNANZA_FASTWEB) && waitinglist.GetCount()) {
		m_nLastStartUpload = ::GetTickCount();
		// Modifica AdunanzA
#if 0
		AddUpNextClient();
#endif
		//
		// Solo quando devo scegliere un client
		// a caso scelgo uno secondo "AduGetType"
		AddUpNextClient(NULL, AduGetTypeBand());
		// fine mod adu
	}

	// The loop that feeds the upload slots with data.
	POSITION pos = uploadinglist.GetHeadPosition();
	uint64 aduSentBytes = 0, fakeSentBytes = 0, aduFiberSentBytes = 0;

	while (pos != NULL) {
		// Get the client. Note! Also updates pos as a side effect.
		CUpDownClient* cur_client = uploadinglist.GetNext(pos);
		//It seems chatting or friend slots can get stuck at times in upload.. This needs looked into..
		if (!cur_client->GetSocket()) {
			RemoveFromUploadQueue(cur_client, true);
			if(cur_client->Disconnected(_T("CUploadQueue::Process"))){
				cur_client->Safe_Delete();
			}
		} else {
			uint32 sentBytes = cur_client->SendBlockData();

			if (cur_client->IsAduClient()) {
				aduSentBytes += sentBytes;
				if (cur_client->IsAduFiber()) 
					aduFiberSentBytes += sentBytes;
			}
			fakeSentBytes += sentBytes;
		}
	}

	// Save used bandwidth for speed calculations
	uint64 sentBytes = theApp.uploadBandwidthThrottler->GetNumberOfSentBytesSinceLastCallAndReset();
	(void)theApp.uploadBandwidthThrottler->GetNumberOfSentBytesOverheadSinceLastCallAndReset();

	// Update statistics
	if (sentBytes) {
		theStats::AddSentBytes(fakeSentBytes,aduSentBytes,aduFiberSentBytes);
	}
}

bool CUploadQueue::AcceptNewClient(bool isAduClient)
{

	uint32 typeClient = AduNextClient();

	switch (typeClient) {
		case ADUNANZA_NONE:
			return false;
			break;
		case ADUNANZA_FASTWEB:
			if (!isAduClient)
				return false;
			break;
		case ADUNANZA_EXTERN:
			if (isAduClient)
				return false;
			break;
	}

	// check if we can allow a new client to start downloading from us
	if (::GetTickCount() - m_nLastStartUpload < 1000 || uploadinglist.GetCount() >= MAX_UP_CLIENTS_ALLOWED) {
		return false;
	}

	float kBpsUpPerClient = (float)thePrefs::GetSlotAllocation();
	float kBpsUp = theStats::GetUploadRate() / 1024.0f;
	if (thePrefs::GetMaxUpload() == UNLIMITED) {
		if ((uint32)uploadinglist.GetCount() < ((uint32)((kBpsUp)/kBpsUpPerClient)+2)) {
			return true;
		}
	} else {
		uint16 nMaxSlots = AduGetMaxUploadSlots();
#if 0
		if (thePrefs::GetMaxUpload() >= 10) {
			nMaxSlots = (uint16)floor((float)thePrefs::GetMaxUpload() / kBpsUpPerClient + 0.5);
				// floor(x + 0.5) is a way of doing round(x) that works with gcc < 3 ...
			if (nMaxSlots < MIN_UP_CLIENTS_ALLOWED) {
				nMaxSlots=MIN_UP_CLIENTS_ALLOWED;
			}
		} else {
			nMaxSlots = MIN_UP_CLIENTS_ALLOWED;
		}
#endif

		//AddDebugLogLineM(false, logClient, CFormat(wxT("in up: %u massimi: %u")) % (uint32)uploadinglist.GetCount() % nMaxSlots);

		if ((uint32)uploadinglist.GetCount() < nMaxSlots) {
			return true;
		}
	}
	return false;
}

CUploadQueue::~CUploadQueue()
{
	wxASSERT(waitinglist.IsEmpty());
	wxASSERT(uploadinglist.IsEmpty());
}

POSITION CUploadQueue::GetWaitingClient(CUpDownClient* client)
{
	return waitinglist.Find(client); 
}

CUpDownClient* CUploadQueue::GetWaitingClientByIP(uint32 dwIP)
{
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)) {
		if (dwIP == waitinglist.GetAt(pos)->GetIP()) {
			return waitinglist.GetAt(pos);
		}
	}
	return 0;
}

CUpDownClient* CUploadQueue::GetWaitingClientByIP_UDP(uint32 dwIP, uint16 nUDPPort)
{
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)) {
		CUpDownClient* cur_client = waitinglist.GetAt(pos);
		if ((dwIP == cur_client->GetIP()) && (nUDPPort == cur_client->GetUDPPort())) {
			return waitinglist.GetAt(pos);
		}
	}
	return 0;
}

POSITION CUploadQueue::GetDownloadingClient(CUpDownClient* client)
{
	return uploadinglist.Find( client );
}

void CUploadQueue::AddClientToQueue(CUpDownClient* client)
{
// Mod Adu
// Emanem
// Questa mod serve a supportare piu' di 50 clients se
// connessi con un low-id
#if 0
	if (theApp.serverconnect->IsConnected() && theApp.serverconnect->IsLowID() && !theApp.serverconnect->IsLocalServer(client->GetServerIP(),client->GetServerPort()) && client->GetDownloadState() == DS_NONE && !client->IsFriend() && theStats::GetWaitingUserCount() > 50) {
#endif
	if (theApp.serverconnect->IsConnected() && theApp.serverconnect->IsLowID() && !theApp.serverconnect->IsLocalServer(client->GetServerIP(),client->GetServerPort()) && client->GetDownloadState() == DS_NONE && !client->IsFriend() && theStats::GetWaitingUserCount() > theApp.glob_prefs->GetQueueSize() ) {
	// Well, all that issues finish in the same: don't allow to add to the queue
		return;
	}

	// fine mod Adu


	if ( client->IsBanned() ) {
		return;
	}

	client->AddAskedCount();
	client->SetLastUpRequest();

	// Find all clients with the same user-hash
	CClientList::SourceList found = theApp.clientlist->GetClientsByHash( client->GetUserHash() );

	CClientList::SourceList::iterator it = found.begin();
	for ( ; it != found.end(); it++ ) {
		if ( IsOnUploadQueue( *it ) ) {
			CUpDownClient* cur_client = *it;

			if ( cur_client == client ) {
				// Stefano Picerno: aggiunto && (uploadinglist.GetCount() < AduGetMaxUploadSlots())
#if 0
				if ( client->m_bAddNextConnect && ( ( uploadinglist.GetCount() < thePrefs::GetMaxUpload() ) || ( thePrefs::GetMaxUpload() == UNLIMITED ) ) ) {
#endif
				//AddDebugLogLineM(false, logClient, CFormat(wxT("in up: %u massimi: %u")) % (uint32)uploadinglist.GetCount() % AduGetMaxUploadSlots());

				if ( client->m_bAddNextConnect && ( ( uploadinglist.GetCount() < thePrefs::GetMaxUpload() ) || ( thePrefs::GetMaxUpload() == UNLIMITED ) ) && (uploadinglist.GetCount() < AduGetMaxUploadSlots()) ) {
					if (lastupslotHighID) {
						client->m_bAddNextConnect = false;
						RemoveFromWaitingQueue(client, true);
						AddUpNextClient(client);
						lastupslotHighID = false; // LowID alternate
						return;
					}
				}

				client->SendRankingInfo();
				Notify_QlistRefreshClient(client);
				return;

			} else {
				// Hash-clash, remove unidentified clients (possibly both)
				
				if ( !cur_client->IsIdentified() ) {
					// Cur_client isn't identifed, remove it
					theApp.clientlist->AddTrackClient( cur_client );

					RemoveFromWaitingQueue( cur_client );
					if ( !cur_client->GetSocket() ) {
						if (cur_client->Disconnected( wxT("AddClientToQueue - same userhash") ) ) {
							cur_client->Safe_Delete();
						}
					}
				}

				if ( !client->IsIdentified() ) {
					// New client isn't identified, remove it
					theApp.clientlist->AddTrackClient( client );

					if ( !client->GetSocket() ) {
						if ( client->Disconnected( wxT("AddClientToQueue - same userhash") ) ) {
							client->Safe_Delete();
						}
					}

					return;
				}
			}
		}
	}

	// Count the number of clients with the same IP-address
	found = theApp.clientlist->GetClientsByIP( client->GetIP() );

	int ipCount = 0;
	for ( it = found.begin(); it != found.end(); it++ ) {
		if ( ( *it == client ) || IsOnUploadQueue( *it ) ) {
			ipCount++;
		}
	}

	// We do not accept more than 3 clients from the same IP
	if ( ipCount > 3 ) {
		return;
	} else if ( theApp.clientlist->GetClientsFromIP(client->GetIP()) >= 3 ) {
		return;
	}

	// statistic values
	CKnownFile* reqfile = (CKnownFile*) client->GetUploadFile();
	if (reqfile) {
		reqfile->statistic.AddRequest();
	}

	// TODO find better ways to cap the list

	//AddDebugLogLineM(false, logClient, CFormat(wxT("in coda: %u massimi: %u")) % (uint32)waitinglist.GetCount() % thePrefs::GetQueueSize());

	if ( (uint32)waitinglist.GetCount() >= (thePrefs::GetQueueSize())) {
		return;
	}

	if (client->IsDownloading()) {
		// he's already downloading and wants probably only another file
		CPacket* packet = new CPacket(OP_ACCEPTUPLOADREQ,0);
		theStats::AddUpOverheadFileRequest(packet->GetPacketSize());
		client->SendPacket(packet,true);
		return;
	}

	if (waitinglist.IsEmpty() && AcceptNewClient(client->IsAduClient())) {
		AddUpNextClient(client);
		m_nLastStartUpload = ::GetTickCount();
	} else {
		waitinglist.AddTail(client);
		theStats::AddWaitingClient();
		client->ClearWaitStartTime();
		client->ClearAskedCount();
		// mod Adu
		// Emanem
		// se un client e' Adu incremento il conteggio
		if (client->IsAduClient()) 
			m_AduClientsNum++;
		// fine mod Adu
		client->SetUploadState(US_ONUPLOADQUEUE);
		client->SendRankingInfo();
		Notify_QlistAddClient(client);
		Notify_ShowQueueCount(waitinglist.GetCount());
	}
}

bool CUploadQueue::RemoveFromUploadQueue(CUpDownClient* client, bool updatewindow)
{
	theApp.clientlist->AddTrackClient(client); // Keep track of this client
	POSITION pos = uploadinglist.Find( client );
	if ( pos != NULL ) {
		if (updatewindow) {
			Notify_UploadCtrlRemoveClient(client);
		}
		uploadinglist.RemoveAt(pos);
		// Mod Adu
		// lupz
		POSITION adupos=aduuploadinglist.Find(client);
		if (adupos)
			aduuploadinglist.RemoveAt(adupos);
		// Fine mod Adu
		theStats::RemoveUploadingClient();
		if( client->GetTransferredUp() ) {
			theStats::AddSuccessfulUpload();
			theStats::AddUploadTime(client->GetUpStartTimeDelay() / 1000);
		} else {
			theStats::AddFailedUpload();
		}
		client->SetUploadState(US_NONE);
		client->ClearUploadBlockRequests();
		return true;
	}
	return false;
}

bool CUploadQueue::CheckForTimeOver(CUpDownClient* client)
{
	// Mod Adu
	// Emanem
	// Se il client e' AdunanzA lo si lascia connesso
	// al max per 1 ora. Se e' esterno al max per
	// 10 minuti.
	if ( client && client->IsAduClient() )
	{
		if( client->GetUpStartTimeDelay() > SESSIONMAXTIME ){ // Try to keep the clients from downloading for ever.
			if (thePrefs::GetVerbose())
				AddDebugLogLineM(false, logClient, wxString::Format(_("%s: Upload session ended due to max time %s."), client->GetUserName().c_str(), CastSecondsToHM(SESSIONMAXTIME/1000).c_str() ));
			return true;
		}
	}
	else
	{
		if( client->GetUpStartTimeDelay() > MIN2MS(10) ) { // Try to keep the clients from downloading for ever.
			return true;
		}
	}
	// Fine mod Adu

	if (thePrefs::TransferFullChunks()) {
		// Allow the client to download a specified amount per session
		if ( client->IsAduClient() )
		{
			/* Kaiser -- 08/04/2004 - 15.43
			   Se è un ext lo "stacca" dopo SESSIONMAXTRANS (9,3mb) altrimenti cerca
			   di staccarlo dopo ADU_SESSIONMAXTRANS (18,6mb) ovvero il doppio.
			   In questo modo possiamo favorire lo scambio tra aduner senza dover 
			   modificare i tempi di reask e rischiando quindi meno ban! */
			if( client->GetQueueSessionPayloadUp() > ADU_SESSIONMAXTRANS){
				// Inoltre se la prossima slot deve essere AdunanzA e non ho altri Adu in coda
				// non lo butto giu' per poi riprenderlo in coda di upload.
				// Quindi se la prossima slot e' non adunanza lo butto giu'.
				// Condizioni sufficiente per buttare giu' un client sono:
				// - Avere altri Adu in coda di attesa
				// - Che il prossimo client non sia Adu
				if ((theApp.uploadqueue->GetAdunanzAUserCount() > 0) || (ADUNANZA_FASTWEB != AduGetTypeBand()))
				{
					if (thePrefs::GetVerbose())
						AddDebugLogLineM(false, logClient,  wxString::Format(_("%s: Upload session ended due to max transfered amount. %s"), client->GetUserName().c_str(), CastItoXBytes(ADU_SESSIONMAXTRANS).c_str() ) );
					return true;
				}
			}
		}
		else
		{
			if( client->GetQueueSessionPayloadUp() > SESSIONMAXTRANS || AduMaxTrans() ){
				if (thePrefs::GetVerbose())
					AddDebugLogLineM(false, logClient, wxString::Format(_("%s: Upload session ended due to max transfered amount. %s"), client->GetUserName().c_str(), CastItoXBytes(SESSIONMAXTRANS).c_str() ));
				return true;
			}
		}
	} else {
		// Stefano Picerno: emule adunanza usa sempre TransferFullChunks = true !!!!!
		AddDebugLogLineM(false, logGeneral, _("ERROR: aMule adunanza assumes that TransferFullChunks is TRUE, but now it's false !!!." ) );
		// Fine modifica
	}
	return false;
}

// Stefano Picerno: Metodo mancante
void CUploadQueue::UpdateMaxClientScore()
{
	m_imaxscore=0;
	// Inizio mod Adu
	m_AduImaxscore = 0;
	m_ExtImaxscore = 0;
	for(POSITION pos = waitinglist.GetHeadPosition(); pos != 0; ) {
		CUpDownClient*	client = waitinglist.GetNext(pos);
		uint32 score = client->GetScore(true, false);
		if(score > m_imaxscore ) 
			m_imaxscore=score;

		if (client->IsAduClient())
			if (score > m_AduImaxscore) 
				m_AduImaxscore = score;
		else 
			if (score > m_ExtImaxscore) 
				m_ExtImaxscore = score;
	}
	// fine mod Adu
}
// Fine

uint16 CUploadQueue::GetWaitingPosition(CUpDownClient* client)
{
	if ( !IsOnUploadQueue(client) ) {
		return 0;
	}

	uint16 rank = 1;
	uint32 myscore = client->GetScore(false);
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0; ) {
		if (waitinglist.GetNext(pos)->GetScore(false) > myscore) {
			rank++;
		}
	}
	return rank;
}

/*
 * This function removes a file indicated by filehash from suspended_uploads_list.
 */
void CUploadQueue::ResumeUpload( const CMD4Hash& filehash )
{
	//Find the position of the filehash in the list and remove it.
	suspendlist::iterator it = std::find( suspended_uploads_list.begin(), 
			                              suspended_uploads_list.end(),
			                              filehash );
	if ( it != suspended_uploads_list.end() )
		suspended_uploads_list.erase( it );
	
	AddLogLineM( false, CFormat( _("Resuming uploads of file: %s" ) )
				% filehash.Encode() );
}

/*
 * This function adds a file indicated by filehash to suspended_uploads_list
 */
void CUploadQueue::SuspendUpload( const CMD4Hash& filehash )
{
	AddLogLineM( false, CFormat( _("Suspending upload of file: %s" ) )
				% filehash.Encode() );

	//Append the filehash to the list.
	suspended_uploads_list.push_back(filehash);
	wxString base16hash = filehash.Encode();

	POSITION pos = uploadinglist.GetHeadPosition();
	while(pos) { //while we have a valid position
		CUpDownClient *potential = uploadinglist.GetNext(pos);
		//check if the client is uploading the file we need to suspend
		if(potential->GetUploadFileID() == filehash) {
			//remove the unlucky client from the upload queue and add to the waiting queue
			RemoveFromUploadQueue(potential, true);

			waitinglist.AddTail(potential);
			theStats::AddWaitingClient();
			potential->SetUploadState(US_ONUPLOADQUEUE);
			potential->SendRankingInfo();
			Notify_QlistRefreshClient(potential);
			Notify_ShowQueueCount(waitinglist.GetCount());
		}
	}
}

bool CUploadQueue::RemoveFromWaitingQueue(CUpDownClient* client, bool updatewindow)
{
	POSITION pos = waitinglist.Find(client);
	if (pos) {
		RemoveFromWaitingQueue(pos);
		if (updatewindow) {
			Notify_ShowQueueCount(waitinglist.GetCount());
		}
		return true;
	} else {
		return false;
	}
}


void CUploadQueue::RemoveFromWaitingQueue(POSITION pos)
{
	CUpDownClient* todelete = waitinglist.GetAt(pos);
	// mod Adu
	// Emanem
	// se il client era adu decremento il contatore
	if (todelete->IsAduClient()) 
		m_AduClientsNum--;
	// fine mod Adu
	waitinglist.RemoveAt(pos);
	theStats::RemoveWaitingClient();
	if( todelete->IsBanned() ) {
		todelete->UnBan();
	}
	Notify_QlistRemoveClient(todelete);
	todelete->SetUploadState(US_NONE);
}


CUpDownClient* CUploadQueue::GetNextClient(CUpDownClient* lastclient)
{
	if (waitinglist.IsEmpty()) {
		return 0;
	}
	if (!lastclient) {
		return waitinglist.GetHead();
	}
	POSITION pos = waitinglist.Find(lastclient);
	if (!pos) {
		return waitinglist.GetHead();
	}
	waitinglist.GetNext(pos);
	if (!pos) {
		return NULL;
	} else {
		return waitinglist.GetAt(pos);
	}
}
